﻿using System;
using System.Threading;
using NoxService.NWS;
using System.Threading.Tasks;

namespace WalletLibrary
{
  public abstract class AsyncTickProcessor
  {
    Task GetTask(CancellationToken cancellationToken, Func<Task<WorkResponse>> Work)
    {
      TaskCompletionSource<WPCompletedEventArgs> ntask = new TaskCompletionSource<WPCompletedEventArgs>();
      TaskCompleted += (e) => CompletedEvent(ntask, e, () => e);

      cancellationToken.Register(() =>
      {
        ntask.TrySetCanceled();
      });

      Work.Invoke();
      return ntask.Task;
    }
    void CompletedEvent<T>(TaskCompletionSource<T> tcs, System.ComponentModel.AsyncCompletedEventArgs e, Func<T> getResult)
    {
      try
      {
        if (e.Error != null) tcs.TrySetException(e.Error); else if (e.Cancelled) tcs.TrySetCanceled(); else tcs.TrySetResult(getResult());
      }
      catch (Exception)
      {
        tcs.TrySetCanceled();
      }
    }

    delegate void WPCompletedEventHandler(WPCompletedEventArgs e);

    event WPCompletedEventHandler TaskCompleted;
    protected virtual void HandleWorkResponse(WorkResponse data)
    {
      if (TaskCompleted == null) return;

      TaskCompleted.Invoke(new WPCompletedEventArgs(data));
    }
    private async Task Process(Func<Task<WorkResponse>> Work)
    {
      if (Work == null) return;

      await GetTask(cancellation.Token, Work);
    }

    readonly object _SignalLock = new object();

    protected CancellationTokenSource cancellation;

    Device noxDevice; readonly HWS hWS = new HWS();

    Task _WorkThread; bool _ProcessData;

    protected AsyncTickProcessor(Device _noxDevice)
    {
      noxDevice = _noxDevice;
    }
    protected virtual bool Start()
    {
      lock (_SignalLock)
      {
        if (_WorkThread != null && !_WorkThread.IsCompleted) return false;

        if (_ProcessData) return false;

        _ProcessData = true;

        _WorkThread = ResponseTask();

        return true;
      }
    }
    protected virtual bool Stop()
    {
      lock (_SignalLock)
      {
        if (_WorkThread == null) return false;

        if (_WorkThread.IsCompleted) return false;

        if (cancellation.IsCancellationRequested) return false;

        if (!_ProcessData) return false;

        cancellation.Cancel();
        return true;
      }
    }
    protected virtual void AddDelay(double time_seconds = 30)
    {
      cancellation.CancelAfter(TimeSpan.FromSeconds(time_seconds));
    }


    TimeSpan REQUEST_TIMEOUT_MSEC = TimeSpan.FromSeconds(30);

    TimeSpan LOCALL_CONNECTIVITY_TIMEOUT_MSEC = TimeSpan.FromSeconds(10);

    TimeSpan CAPTIVE_RATE_TIMEOUT_MSEC = TimeSpan.FromSeconds(10);

    TimeSpan ADD_DELAY_TIMEOUT_MSEC = TimeSpan.FromSeconds(15);

    TimeSpan CAPTIVE_RATE_PAUSE_MSEC = TimeSpan.FromMilliseconds(200);

    TimeSpan REQUEST_PAUSE_MSEC = TimeSpan.FromMilliseconds(500);

    private async Task ResponseTask()
    {

      WorkResponse ExceptionResponse = null;

      try
      {
        cancellation = new CancellationTokenSource(LOCALL_CONNECTIVITY_TIMEOUT_MSEC);
        cancellation.Token.ThrowIfCancellationRequested();

        ExceptionResponse = new WorkResponse() { ValidationRate = "Disconnected", Action = ResponseActionType.NOWIFI };

        await Process(_ConnectionTicks());

        cancellation.CancelAfter(CAPTIVE_RATE_TIMEOUT_MSEC);

        ExceptionResponse = new WorkResponse() { ValidationRate = "connected but no internet", Action = ResponseActionType.NOWIFI };

        await Process(_RateTicks());

        ExceptionResponse = new WorkResponse() { CancellationRequested = true, Action = ResponseActionType.CLOSE };

        cancellation.CancelAfter(REQUEST_TIMEOUT_MSEC);

        await Process(_WSTicks());

      }
      catch (TaskCanceledException)
      {
        HandleWorkResponse(ExceptionResponse);
      }
      catch (Exception)
      {
        HandleWorkResponse(new WorkResponse() { CancellationRequested = true, Action = ResponseActionType.CLOSE });
      }
      finally
      {
        _ProcessData = false;
      }
    }
    protected virtual Func<Task<WorkResponse>> _ConnectionTicks()
    {
      return null;
    }
    Func<Task<WorkResponse>> _RateTicks()
    {
      return (async () =>
      {

        bool nowifi = true;

        while (nowifi && !cancellation.IsCancellationRequested)
        {
          string rate = await NoxService.NXSC.NXSC.GetRateSpot();

          if (string.IsNullOrEmpty(rate))
          {
            await Task.Delay(CAPTIVE_RATE_PAUSE_MSEC);
          }
          else
          {
            nowifi = false;
            var resp = new WorkResponse() { ValidationRate = "1 BTC = " + rate + " USD", Action = ResponseActionType.SHOW_RATE };
            HandleWorkResponse(resp);
            return resp;
          }
        }

        return new WorkResponse() { Action = ResponseActionType.NOWIFI };
      });
    }

    Func<Task<WorkResponse>> _WSTicks()
    {
      return (async () =>
      {
        while (!cancellation.IsCancellationRequested)
        {
          var resp = await Request();

          if (resp != null)
          {
            var workResponse = new WorkResponse() { UserInfo = resp, Action = ResponseActionType.REQUEST };
            HandleWorkResponse(workResponse);
            return workResponse;
          }

          await Task.Delay(REQUEST_PAUSE_MSEC);
        }

        return new WorkResponse() { Action = ResponseActionType.NONE };
      });
    }
    async Task<FormUserInfo> Request()
    {
      try
      {
        var hWSResponse = await hWS.GetCurrentSMSToken(noxDevice, cancellation.Token);

        if (hWSResponse.Notfify) return hWSResponse.formUserInfo;
      }
      catch { }
      return null;
    }

    class WPCompletedEventArgs : System.ComponentModel.AsyncCompletedEventArgs
    {
      WorkResponse _obj;
      public WPCompletedEventArgs(WorkResponse result) : base(null, false, null)
      {
        _obj = result;
      }
      public WorkResponse GetResult()
      {
        return _obj;
      }
    }
    protected enum ResponseActionType
    {
      NONE = 0,
      REQUEST = 1,
      SHOW_RATE = 2,
      NOWIFI = 3,
      CLOSE = 4
    }
    protected class WorkResponse
    {
      public FormUserInfo UserInfo { get; set; }
      public bool CancellationRequested { get; set; }
      public string ValidationRate { get; set; }

      public ResponseActionType Action;
    }
  }
}
