using System;
using EthereumLibrary.BouncyCastle.Crypto;
using EthereumLibrary.BouncyCastle.Utilities;
using EthereumLibrary.BouncyCastle.Asn1;
using EthereumLibrary.BouncyCastle.Asn1.X9;

namespace EthereumLibrary.BouncyCastle.Crypto.Parameters
{
	internal abstract class ECKeyParameters
		: AsymmetricKeyParameter
	{
		private static readonly string[] algorithms = { "EC", "ECDSA", "ECDH", "ECDHC", "ECGOST3410", "ECMQV" };

		private readonly string algorithm;
		private readonly ECDomainParameters parameters;

		protected ECKeyParameters(
			string algorithm,
			bool isPrivate,
			ECDomainParameters parameters)
			: base(isPrivate)
		{
			if(algorithm == null)
				throw new ArgumentNullException("algorithm");
			if(parameters == null)
				throw new ArgumentNullException("parameters");

			this.algorithm = VerifyAlgorithmName(algorithm);
			this.parameters = parameters;
		}

		public string AlgorithmName
		{
			get
			{
				return algorithm;
			}
		}

		public ECDomainParameters Parameters
		{
			get
			{
				return parameters;
			}
		}

		public override bool Equals(
			object obj)
		{
			if(obj == this)
				return true;

			ECDomainParameters other = obj as ECDomainParameters;

			if(other == null)
				return false;

			return Equals(other);
		}

		protected bool Equals(
			ECKeyParameters other)
		{
			return parameters.Equals(other.parameters) && base.Equals(other);
		}

		public override int GetHashCode()
		{
			return parameters.GetHashCode() ^ base.GetHashCode();
		}

		internal static string VerifyAlgorithmName(string algorithm)
		{
			string upper = Platform.ToUpperInvariant(algorithm);
			if(Array.IndexOf(algorithms, algorithm, 0, algorithms.Length) < 0)
				throw new ArgumentException("unrecognised algorithm: " + algorithm, "algorithm");
			return upper;
		}

        internal static ECDomainParameters LookupParameters(
           DerObjectIdentifier publicKeyParamSet)
        {
            if (publicKeyParamSet == null)
                throw new ArgumentNullException("publicKeyParamSet");

            ECDomainParameters p = null; // EC ECGost3410NamedCurves.GetByOid(publicKeyParamSet);

            if (p == null)
            {
                X9ECParameters x9 = EC.CustomNamedCurves.Secp256k1; // (publicKeyParamSet);

                if (x9 == null)
                {
                    throw new ArgumentException("OID is not a valid public key parameter set", "publicKeyParamSet");
                }

                p = new ECDomainParameters(x9.Curve, x9.G, x9.N, x9.H, x9.GetSeed());
            }

            return p;
        }
	}
}
